# WhatAmIThinking-JsonProto

Declarative json modeler, validator, encoder, and decoder

## Benchmarks

Please take all benchmarks with a grain of salt. Results depend heavily on what you are doing. Different libraries may be optimized for certain situations, such as simplistic benchmarking approachs (including this library). It is always best to do your own benchmarking for the kinds of data models and validation you expect to encounter in your project. That said, I usually like seeing benchmarks on infrastructure stuff which I will use everywhere as a quick gut check. If you have suggestions on how to improve these benchmarks, please open an issue.

Also worth mentioning I do what I can to use the best available settings for each library which I found in the docs. I am no doubt a bit biased towards making my own library perform well, but I try to avoid doing any kind of "bending over backwards" to make things optimal unless the docs suggest it. Again, everyone is biased and you should do your own benchmarking.

### Config

- Python: 3.13.3 (tags/v3.13.3:6280bb5, Apr  8 2025, 14:47:33) [MSC v.1943 64 bit (AMD64)]
- OS: Windows 10.0.26100
- Memory: 16 GB
- CPU: Intel64 Family 6 Model 142 Stepping 10, GenuineIntel
- Frequency: 2.00 GHz
- Physical Cores: 4
- Logical Cores: 8

### Basic Model Operations

This covers just working with the models themselves (classes decorated with `@jsonproto`) directly and does not cover time the `Codec` would spend performing various operations, including constructing. One of the areas for improvement noted when auditing modeling libraries was that the time just to define the model class was quite large and significantly slowed down the debugger in an IDE when a large number of models (a few hundred in one project) were created, making it cumbersome to debug the code at all. The time spent building the model is a central focus of this library so it can scale well even with large numbers of models.

```text
Libraries:
- pydantic: 2.10.6
- dataclassy: 1.0.1
- attrs: 25.3.0
- pyserde: 0.24.0
- mashumaro: 3.15
- msgspec: 0.19.0
- whatamithinking.jsonproto: 1.0.1

Mutable Unslotted Basic Model Operations
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+
|                          | import (μs) | create (μs) | equal (μs) | order (μs) | repr (us) | hash (us) | getattr (us) | setattr (us) |
+==========================+=============+=============+============+============+===========+===========+==============+==============+
| **standard classes**     | 16.87       | 0.90        | 0.21       | 0.49       | 1.25      | 0.45      | 0.12         | 0.13         |
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+
| **dataclasses**          | 1453.92     | 0.86        | 0.22       | 0.46       | 1.68      | 0.42      | 0.14         | 0.13         |
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+
| **pydantic**             | 478.69      | 9.63        | 3.60       | N/A        | 7.08      | N/A       | 0.23         | 22.01        |
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+
| **pydantic.dataclasses** | 3708.29     | 3.64        | 0.23       | 0.60       | 1.80      | 0.47      | 0.14         | 0.27         |
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+
| **dataclassy**           | 479.00      | 1.02        | 0.55       | 0.78       | 8.84      | 0.29      | 0.15         | 0.13         |
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+
| **attrs**                | 1260.40     | 0.95        | 0.21       | 5.35       | 1.97      | 0.44      | 0.26         | 0.13         |
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+
| **pyserde**              | 7197.45     | 2.99        | 0.22       | 0.56       | 1.89      | 0.45      | 0.13         | 0.13         |
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+
| **mashumaro**            | 4354.11     | 0.90        | 0.22       | 0.46       | 1.72      | 0.42      | 0.14         | 0.14         |
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+
| **msgspec**              | 34.69       | 0.22        | 0.04       | 0.10       | 1.09      | N/A       | 0.12         | 0.38         |
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+
| **jsonproto**            | 35.12       | 1.01        | 0.23       | 0.53       | 1.12      | 0.43      | 0.14         | 0.17         |
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+

Frozen Unslotted Basic Model Operations
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+
|                          | import (μs) | create (μs) | equal (μs) | order (μs) | repr (us) | hash (us) | getattr (us) | setattr (us) |
+==========================+=============+=============+============+============+===========+===========+==============+==============+
| **standard classes**     | 16.92       | 2.12        | 0.21       | 0.55       | 1.24      | 0.41      | 0.14         | N/A          |
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+
| **dataclasses**          | 1868.21     | 2.00        | 0.22       | 0.47       | 1.76      | 0.44      | 0.14         | N/A          |
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+
| **pydantic**             | 477.81      | 3.36        | 3.74       | N/A        | 7.23      | 0.63      | 0.20         | N/A          |
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+
| **pydantic.dataclasses** | 4542.48     | 3.37        | 0.23       | 0.58       | 1.68      | 0.49      | 0.15         | N/A          |
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+
| **dataclassy**           | 571.04      | 2.06        | 0.55       | 0.92       | 8.60      | 0.29      | 0.14         | N/A          |
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+
| **attrs**                | 1315.89     | 1.50        | 0.20       | 5.48       | 2.00      | 0.44      | 0.14         | N/A          |
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+
| **pyserde**              | 7673.07     | 4.11        | 0.23       | 0.47       | 1.69      | 0.41      | 0.14         | N/A          |
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+
| **mashumaro**            | 4814.53     | 2.09        | 0.26       | 0.65       | 1.74      | 0.46      | 0.18         | N/A          |
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+
| **msgspec**              | 164.49      | 0.30        | 0.04       | 0.10       | 1.06      | 0.12      | 0.12         | N/A          |
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+
| **jsonproto**            | 27.55       | 1.91        | 0.20       | 0.47       | 0.24      | 0.33      | 0.14         | N/A          |
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+

Mutable Slotted Basic Model Operations
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+
|                          | import (μs) | create (μs) | equal (μs) | order (μs) | repr (us) | hash (us) | getattr (us) | setattr (us) |
+==========================+=============+=============+============+============+===========+===========+==============+==============+
| **standard classes**     | 18.27       | 1.05        | 0.22       | 0.60       | 1.11      | 0.47      | 0.13         | 0.13         |
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+
| **dataclasses**          | 1518.97     | 0.88        | 0.22       | 0.76       | 1.70      | 0.44      | 0.13         | 0.13         |
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+
| **pydantic**             | N/A         | N/A         | N/A        | N/A        | N/A       | N/A       | N/A          | N/A          |
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+
| **pydantic.dataclasses** | 3750.41     | 3.70        | 0.23       | 0.59       | 1.82      | 0.48      | 0.12         | 0.13         |
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+
| **dataclassy**           | 494.52      | 0.88        | 0.59       | 0.76       | 8.81      | 0.31      | 0.13         | 0.13         |
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+
| **attrs**                | 1376.97     | 0.96        | 0.20       | 10.10      | 3.09      | 0.95      | 0.14         | 0.37         |
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+
| **pyserde**              | 7269.07     | 2.96        | 0.22       | 0.63       | 1.84      | 0.44      | 0.12         | 0.13         |
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+
| **mashumaro**            | 8497.96     | 0.89        | 0.22       | 0.55       | 1.71      | 0.56      | 0.12         | 0.13         |
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+
| **msgspec**              | 38.08       | 0.35        | 0.14       | 0.20       | 1.65      | N/A       | 0.14         | 0.58         |
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+
| **jsonproto**            | 106.34      | 1.04        | 0.30       | 0.61       | 1.25      | 0.45      | 0.13         | 0.18         |
+--------------------------+-------------+-------------+------------+------------+-----------+-----------+--------------+--------------+
```

### Validation and Conversion

```text
Number of executions per trial: 10000
Number of repeated trials: 5 (best time taken)

Benchmarking Pydantic (BaseModel)...
  Valid Data Success:     8.45 µs/op
Benchmarking JsonProto...
  Valid Data Success:     58.75 µs/op
Benchmarking Attrs...
  Valid Data Success:     3.52 µs/op
Benchmarking Pyserde...
  Valid Data Success:     12.99 µs/op
Benchmarking Mashumaro...
  Valid Data Success:     7.02 µs/op
Benchmarking Msgspec...
  Valid Data Success:     1.56 µs/op
Benchmarking Apischema...
  Valid Data Success:     25.97 µs/op
```
